

#ifndef _NewThreads_h_
#define _NewThreads_h_ 1


/*****************************************************************************
 *
 *  first, define some system-dependent macros 
 */

/* Unless told otherwise, assume one-node configuration */
#define OneNode 

/* These entries are system definitions -- for backwards compatibility. */
#ifdef Ipsc2
#define I386      1
#define Use_IntelNX
#endif

#ifdef Ipsc860
#define I860      1
#define Use_IntelNX
#endif

#ifdef DecstationNet
#define Mips      1
#define Use_PVM       1
#define	Use_PVM2
#endif

#ifdef Decstation
#ifndef Mips
#define	Mips	1
#endif
#endif

#ifdef Snake
#ifndef Hppa
#define	Hppa
#endif
#endif

/* The following are Architecture definitions */

#if defined(Mips)
#define	StackGrowsUp	0	
#endif

#ifdef I860
#define	StackGrowsUp	0	
#endif

#ifdef I386
#define       StackGrowsUp    0
#endif

#if defined(Hppa)
#define	StackGrowsUp	1
/*
 * KLUDGE XXX
 *
 * There are multiple precision-architecture versions; so far,
 * they differ only in the number of floating point registers.
 * Unfortunately, I don't know of any way to tell them apart
 * at compilation time.
 *
 * Until I discover how to tell them apart at runtime, just
 * assume that all hppa machines are the newer ones
 */
#ifndef spectrum
#define	spectrum	/* pa risc ver. >1.0 */
#endif

#endif /* Hppa */

#ifdef Sparc
#define	StackGrowsUp	0
#endif /* Sparc */

#ifdef Power
#define	StackGrowsUp	0
#endif /* rs6000 */

#ifdef Mc68000
#define	StackGrowsUp	0
#define	Mc68020		/* undef this if you don't have 680x0 >= 68020 */
#endif /* mc68000 */

/* These are communications facility definitions */ 

#ifdef Use_PVM
#ifdef OneNode
#undef OneNode
#endif
#ifdef Use_PVM2
#define PVM	1
#else
#define	PVM3	1
#endif
#endif

#ifdef Use_IntelNX
#ifdef OneNode
#undef OneNode
#endif
#define	IntelNX
#endif


/*****************************************************************************
 *
 * threads and synchronization 
 */


#define	ThreadStackSize	 (64*1024)

typedef enum { 
  Defunct,   /* uninitialized or dead */
  Virgin,    /* ready, but never been run */
  Ready,     /* ready to run */
  Running,   /* currently running on a CPU */
  Blocked,   /* waiting for something to happen */
} ThreadStatus;


typedef void (*ThreadFunction)(void*);  /* type of a thread's start-function */

typedef struct Thread {
  void*          userStuff;       /* user can do what s/he wants with this */

  void*          saveSp;          /* saved stack-pointer */

  ThreadStatus   status;          

  ThreadFunction startFunc;       /* function that thread should execute */
  void*          startArg;        /* argument for that function */

  struct Thread* readyQ;          /* used for linking the ready-thread queue */
  struct Thread* next;            /* used for other linked lists */

  void*          msgBuf;         /* describe buffer that thread is waiting */
  int            msgSize;        /*   to receive into */

  int            usesFloatingPoint; /* does this thread use floating-point? */

  char           name[64];          /* for debugging only */

  char           stack[ThreadStackSize];  /* the thread's stack */
} Thread;

/* thisThread always points to the currently executing thread */
extern Thread* thisThread;   

/* return version number */
char* newthreadsVersion();

/* threadStart "fires up" a thread.  The thread will start running sometime
 * soon after the call.  The thread will execute the procedure-call func(arg).
 */
void threadStart(Thread* t, ThreadFunction func, void* arg);

/* threadBlock causes the current thread to block indefinitely.  The thread 
 * will not resume until somebody calls threadUnblock on it.  threadUnblock
 * makes a blocked thread runnable again. 
 */
void threadBlock();
void threadUnblock(Thread*);

/* threadYield gives up the CPU so another thread can run.  */
void threadYield();

/* don't call threadHandoffCPU unless you know what you're doing */
void threadHandoffCPU(Thread*, ThreadStatus);

/* threadExit causes the current thread to exit.  The thread will never run
 * again, but it can be re-used by calling threadStart on it again.  Note 
 */
void threadExit();

/* threadUsesFloatingPoint tells the scheduler whether or not the current
 * thread is using the floating-point registers.  (If a thread is not using
 * these registers, it can be context-switched more quickly.)  By default,
 * threads are assumed to be using floating-point.  This procedure can be
 * called as often as desired to turn floating-point register-saving on and
 * off.
 */
void threadUsesFloatingPoint(int usesIt);

void nodeExit();


/* Mutexes are (blocking) locks.  Only one thread at a time may hold a mutex;
 * additional threads trying to acquire the mutex will block until the mutex
 * is free.  
 *
 * A mutex is acquired by calling mutexAcquire, and released by calling
 * mutexRelease.
 *
 * Mutexes must be initialized before use (by calling mutexInit).  They
 * should be destroyed after use (by calling mutexDestroy).
 *
 * Mutexes are not FIFO; a thread trying to acquire the mutex may be infinitely
 * overtaken by other threads acquiring that mutex.  (Fairness has a 
 * performance cost, so we decided to do without it.)
 *
 * The above is the old performance (obtainable with OLD_MUTEX_BEHAVIOUR)
 * The current default is to have mutexs queue requests.
 */
typedef struct {
  int     locked;
  Thread* waiters;
} Mutex;

void mutexInit(Mutex*);
void mutexDestroy(Mutex*);
void mutexAcquire(Mutex*);
void mutexRelease(Mutex*);


/* Conditions are condition-variables with Mesa semantics.  (If you don't 
 * understand what that means, never mind.)  
 * 
 * Conditions must be initialized before use (by calling conditionInit).  They
 * should be destroyed after use (by calling conditionDestroy).
 *
 * Conditions are not FIFO; a thread waiting on a condition may be infinitely
 * overtaken by other threads waiting on that condition.  (Fairness has a 
 * performance cost, so we decided to do without it.)
 *
 * The above is the old performance (obtainable with OLD_CONDITION_BEHAVIOUR)
 * The current default is to have conditions queue requests.
 */

typedef struct {
  Thread* waiters;
} Condition;

void conditionInit(Condition*);
void conditionDestroy(Condition*);
void conditionWait(Condition*, Mutex*);
void conditionSignal(Condition*);
void conditionBroadcast(Condition*);



/*****************************************************************************
 *
 * communication
 */

extern int MyNodeNum;
extern int NumNodes;

void Tsend(void* buf, int size, int nodeNum, int portNum);
int  Trecv(void* buf, int maxSize, int portNum);


/*****************************************************************************
 *
 * spinlocks
 */

#ifdef Multiprocessor

/* nothing implemented yet */
typedef int Spinlock;
#define InitSpinlock(X)      assert(0);
#define AcquireSpinlock(X)   assert(0);
#define ReleaseSpinlock(X)   assert(0);

#else  /* Multiprocessor undefined */

/* These operations are no-ops on a uniprocessor. */
typedef int Spinlock;
#define InitSpinlock(X)      /**/
#define AcquireSpinlock(X)   /**/
#define ReleaseSpinlock(X)   /**/

#endif /* ifdef Multiprocessor */



#endif

