
#include <stdio.h>
#include "newthreads.h"


#ifdef PigsFly
#define Verbose     1
#endif

#define TotalSizeX 32
#define TotalSizeY 32
#define NumWorkers 1

extern char* memcpy();
extern int   atoi(char*);




int numLocalWorkers;
int numWorkers;
int xWorkers, yWorkers;
int totalSizeX, totalSizeY;

typedef double GridPoint;

GridPoint** storedGrid[16];

Thread workers[32];

Mutex countMutex;
Condition allWorkersDone;
int workersLeft;
Thread* userMainThread;


void userMain(int argc, char** argv)
{
  int  i;
  void worker(int);
  GridPoint** getNewGrid(int, int);

  if(argc != 4){
    fprintf(stderr, "usage: %s Xgridsize Ygridsize threadsPerNode\n",
		    argv[0]);
    exit(1);
  }
  totalSizeX = atoi(argv[1]);
  totalSizeY = atoi(argv[2]);
  numLocalWorkers = atoi(argv[3]);
  numWorkers = numLocalWorkers*NumNodes;  
  for(xWorkers=1; (xWorkers*xWorkers)<numWorkers; xWorkers<<=1)  {;}
  yWorkers = numWorkers/xWorkers;
  for(i=0; i<numLocalWorkers; ++i){
    storedGrid[i] = getNewGrid(totalSizeX/xWorkers, totalSizeY/yWorkers);
  }
  mutexInit(&countMutex);
  conditionInit(&allWorkersDone);
  workersLeft = numLocalWorkers;
  userMainThread = thisThread;
  for(i=0; i<numLocalWorkers; ++i){
	threadStart(&workers[i], worker, (void* )(numLocalWorkers*MyNodeNum+i));
  }
  mutexAcquire(&countMutex);
  while(workersLeft > 0){
    conditionWait(&allWorkersDone, &countMutex);
  }
  mutexRelease(&countMutex);
  nodeExit();
}


int xOff[4] = { 0, 0, 1, -1 };
int yOff[4] = { 1, -1, 0, 0 };
int dirPair[4] = { 1, 0, 3, 2 };


void worker(int workerID)
{
  int   i;
  int   myLocalNum;
  int   myX, myY;
  int   hisX, hisY;
  long  startTime, endTime;
  int   outNode[4], outPort[4], inPort[4];
  GridPoint** theGrid;
  void  compute(GridPoint**, int, int, int[], int[], int[]);
  void  null();

  myX = workerID%xWorkers;
  myY = workerID/xWorkers;
  myLocalNum = workerID % numLocalWorkers;
/*
  myX = (int)gray(myX);
  myY = (int)gray(myY);
*/
#ifdef Verbose
  cout << "(" << myX << " " << myY << ") starting\n";
  cout.flush();
#endif
  for(i=0; i<4; ++i){
	inPort[i] = 4*myLocalNum+i;
	hisX = (myX+xOff[i]+xWorkers)%xWorkers;
	hisY = (myY+yOff[i]+yWorkers)%yWorkers;
	outNode[i] = (hisY*xWorkers+hisX) / numLocalWorkers;
	outPort[i] = 4*((hisY*xWorkers+hisX)%numLocalWorkers)+dirPair[i];
  }

  printf("%d: node %d local %d  X %d Y %d   in %d %d %d %d  out %d %d %d %d\n",
         workerID, MyNodeNum, myLocalNum, myX, myY,
		 inPort[0], inPort[1], inPort[2], inPort[3],
		 outPort[0], outPort[1], outPort[2], outPort[3]);
  fflush(stdout);

  theGrid = storedGrid[workerID%numLocalWorkers];

  compute(theGrid, myX, myY, inPort, outNode, outPort);
  if(workerID == 0){
    time(&startTime);
  }

  for(i=0; i<10; ++i){
    compute(theGrid, myX, myY, inPort, outNode, outPort);
  }
  if(workerID == 0){
    time(&endTime);
	printf("GRID    gridsize %d %d   nodes %d   threads %d   time %lf\n",
		   totalSizeX, totalSizeY, NumNodes, numLocalWorkers, 
		   (endTime-startTime)*1.0);
	fflush(stdout);
  }

  mutexAcquire(&countMutex);
  if(--workersLeft <= 0){
    conditionSignal(&allWorkersDone);
  }
  mutexRelease(&countMutex);
  threadExit();
}



void  compute(GridPoint **theGrid, 
              int myX, int myY, int inPort[], int outNode[], int outPort[])
{
  int i, j;
  int iter;
  int xLocal, yLocal;
  GridPoint *nBuf;
  GridPoint *sBuf;
  GridPoint *eBuf;
  GridPoint *wBuf;

  xLocal = totalSizeX/xWorkers;
  yLocal = totalSizeY/yWorkers;
  
  nBuf = (GridPoint*) malloc(xLocal*sizeof(GridPoint));
  sBuf = (GridPoint*) malloc(xLocal*sizeof(GridPoint));
  eBuf = (GridPoint*) malloc(yLocal*sizeof(GridPoint));
  wBuf = (GridPoint*) malloc(yLocal*sizeof(GridPoint));
  for(iter=0; iter<2; ++iter){
    for(i=0; i<yLocal; ++i){
      eBuf[i] = theGrid[xLocal-1][i];
      wBuf[i] = theGrid[0][i];
    }
    Tsend(eBuf, yLocal*sizeof(GridPoint), outNode[2], outPort[2]);
    Tsend(wBuf, yLocal*sizeof(GridPoint), outNode[3], outPort[3]);
    for(i=0; i<xLocal; ++i){
      nBuf[i] = theGrid[i][yLocal-1];
      sBuf[i] = theGrid[i][0];
    }
    Tsend(nBuf, xLocal*sizeof(GridPoint), outNode[0], outPort[0]);
    Tsend(sBuf, xLocal*sizeof(GridPoint), outNode[1], outPort[1]);
	Trecv(nBuf, xLocal*sizeof(GridPoint), inPort[0]);
	Trecv(sBuf, xLocal*sizeof(GridPoint), inPort[1]);
    for(i=0; i<xLocal; ++i){
      theGrid[i][yLocal] = nBuf[i];
      theGrid[i][-1] = sBuf[i];
    }
	Trecv(eBuf, yLocal*sizeof(GridPoint), inPort[2]);
	Trecv(wBuf, yLocal*sizeof(GridPoint), inPort[3]);
    for(i=0; i<yLocal; ++i){
      theGrid[xLocal][i] = eBuf[i];
      theGrid[-1][i] = wBuf[i];
    }

    for(i=0; i<xLocal; ++i){
      for(j=0; j<yLocal; ++j){
        theGrid[i][j] = (theGrid[i][j] + theGrid[i-1][j] + theGrid[i+1][j] +
                         theGrid[i][j-1] + theGrid[i][j+1])/5.0;
      }
    }
  }
  free(nBuf);
  free(sBuf);
  free(eBuf);
  free(wBuf);
}



GridPoint **getNewGrid(int xLocal, int yLocal)
{
  int i, j;
  GridPoint** mal;
  GridPoint** ret;
  GridPoint* mal2;

  mal = (GridPoint**)malloc((xLocal+2)*sizeof(GridPoint*));
  ret = mal+1;
  for(i=(-1); i<=xLocal; ++i){
	mal2 = (GridPoint*)malloc((yLocal+2)*sizeof(GridPoint));
    ret[i] = mal2+1;
  }
  for(i=0; i<xLocal; ++i){
    for(j=0; j<yLocal; ++j){
      ret[i][j] = 0.0;
    }
  }
  return ret;
}


void null()
{
}

