
#include <stdio.h>
#include <assert.h>
#include "newthreads.h"

#ifndef USEFP
#define	USEFP	0
#endif

#define NumThreads 5

#if defined(SLOW)
/* Because context switches take a long time on some boxes, don't do as many */
#define PongTimes  25000
#define LocalMsgPongTimes 250000
#define RemoteMsgPongTimes 25000
#else
#define PongTimes  1000000
#define LocalMsgPongTimes 1000000
#define RemoteMsgPongTimes 100000
#endif


int ack[NumThreads];
Thread worker[NumThreads];
Thread pongThread;
Thread msgPongThread;

Mutex  theBall;
int    whoseShot;
Condition paddle[2];


void main(int argc, char* argv[])
{
  int i;
  int startTime, endTime;
  void workerFunc(void*);
  void pingPongFunc(void*);
  void msgPongFunc(void*);
  void remotePongFunc(void*);

  init_newthreads(&argc, argv);

  /* print out the arguments */
  printf("%d: %d args\n", MyNodeNum, argc);
  for(i=0; i<argc; ++i){
	printf("%d: arg[%d] = %s\n", MyNodeNum, i, argv[i]);
  }

  /* print some stuff */
  for(i=0; i<NumThreads; ++i){
    ack[i] = 0;
	sprintf(worker[i].name, "worker[%d]", i);
    threadStart(&worker[i], workerFunc, (void*)i);
  }	
  for(i=0; i<NumThreads; ++i){
	while(ack[i] == 0)    threadYield();
  }

  /* play some pong */
  whoseShot = 0;
  mutexInit(&theBall);
  conditionInit(&paddle[0]);  
  conditionInit(&paddle[1]);
  sprintf(pongThread.name, "pongThread");
  threadStart(&pongThread, pingPongFunc, (void*)1);
  startTime = time(0);
  pingPongFunc((void*)0);
  endTime = time(0);
  if(MyNodeNum == 0){
	printf("%d usec per ping.\n", 
		   (int)((500000.0*(endTime-startTime))/PongTimes));
	fflush(stdout);
  }

  /* play some local message-pong */
  threadStart(&msgPongThread, msgPongFunc, (void*)1);
  startTime = time(0);
  msgPongFunc((void*)0);
  endTime = time(0);
  if(MyNodeNum == 0){
	printf("%d usec per local message.\n", 
		   (int)((500000.0*(endTime-startTime))/LocalMsgPongTimes));
	fflush(stdout);
  }
  if( (NumNodes>1) && (MyNodeNum<2) ){
	startTime = time(0);
	remotePongFunc(0);
	endTime = time(0);
	if(MyNodeNum == 0){
	  printf("%d usec per remote message.\n", 
		 (int)((500000.0*(endTime-startTime))/RemoteMsgPongTimes));
	fflush(stdout);
	}
  }
  nodeExit();
}


void workerFunc(void* arg)
{
  int id = (int)arg;
  printf("Hello, world from %d\n", id);
  fflush(stdout);
  ack[id] = 1;
  threadExit();
}


void pingPongFunc(void* arg)
{
  int i;
  int me = (int)arg;

  threadUsesFloatingPoint(USEFP);
  mutexAcquire(&theBall);
  for(i=0; i<PongTimes; ++i){
	while(whoseShot != me){
	  conditionWait(&paddle[me], &theBall);
	}
	whoseShot = 1-me;
	conditionSignal(&paddle[1-me]);
  }
  mutexRelease(&theBall);
}


void msgPongFunc(void* arg)
{
  int i;
  int ret;
  int me = (int)arg;
  int inBuf, outBuf;

  threadUsesFloatingPoint(USEFP);
  outBuf = me;
  for(i=0; i<LocalMsgPongTimes; ++i){
	Tsend(&outBuf, sizeof(int), MyNodeNum, 1-me);
	inBuf = -1;
	ret = Trecv(&inBuf, sizeof(int), me);
	assert(ret == sizeof(int));
	assert(inBuf == 1-me);
  }
}


void remotePongFunc(void* ignoredArg)
{
  int i;
  int ret;
  int me = MyNodeNum;
  int inBuf, outBuf;

  threadUsesFloatingPoint(USEFP);
  outBuf = me;
  for(i=0; i<RemoteMsgPongTimes; ++i){
	Tsend(&outBuf, sizeof(int), 1-me, 2);
	inBuf = -1;
	ret = Trecv(&inBuf, sizeof(int), 2);
	assert(ret == sizeof(int));
	assert(inBuf == 1-me);
  }
}


